#include "hillclimber.h"

extern long TotalNumSubProbs;
extern long BoundNoSolCnt;
extern long SuccGreedyCnt;
extern long SuccLSCnt;
extern long SuccEulerCnt;
extern long FitNoSolCnt;

extern int subProbMethod;
extern int verbosity;
extern int algType;
extern int localSearchFlag;
extern int orderHeuristic;
extern int repairHeuristic;

extern int numItems;
extern int numTypes;
extern int tMin;
extern vector<SHAPE> shape;
extern vector<TYPE> types;

extern ofstream binTimeStream;
extern ofstream binItsStream;
extern ofstream costTimeStream;
extern ofstream costItsStream;
extern ofstream divStream;
extern ofstream resultsLog;

//-------------------------------------------------------------------------------------
void applySearch(vector<vector<PIECE> > &candSol, vector<long> &stockUsage, vector<vector<PIECE> > &unplaced, vector<PIECE> &finalStock, vector<int> &groupsToRemove)
{
	//error check
	if(candSol.size() <= 2){cout<<"Fewer than 3 groups are being used. Aborting...\n"; 	exit(1);}
	int i,j;
	bool changedSomething = false;
	
	for(j=0;j<localSearchFlag;j++){
		selectGroupsToRemove(groupsToRemove,candSol.size());
		//We now copy all of the groups we plan to remove from the candidate solution into a second 2d vector 
		//called "unplaced". We then empty the groups that we wanted to remove, and remove these empty groups
		unplaced.clear();
		unplaced.resize(groupsToRemove.size());
		for(i=0; i<groupsToRemove.size(); i++){
			unplaced[i].swap(candSol[groupsToRemove[i]]);
			stockUsage[groupsToRemove[i]] = 0;
		}
		removeAnyEmptyGroups(candSol,stockUsage);

		//We are now ready to apply the LS procedure
		do 
			changedSomething = runLS(candSol,unplaced,stockUsage,finalStock);
		while(changedSomething == true);

		//Once we are here, we are able to rebuild into a full solution. 
		for(i=0; i<unplaced.size(); i++) {
			stockUsage.push_back(calcStockUsage(unplaced[i]));
			candSol.push_back(unplaced[i]);
		}
	}
	//When we get here (whether having run HS or not) we should have a full solution again. We therefore
	//permute the groups and run greedyInsert
	rebuild(candSol,stockUsage,finalStock);
}	

//-------------------------------------------------------------------------------------
void runHillClimber(int targetGroups, bool usingTL, bool usingIL, int itLimit, int timeLimit, char *filename){

	if(verbosity>=1)
		cout<<"Starting Experiment with HillClimber; File = "<<filename<<"; SubProblemMethod = "<<subProbMethod<<"; LocalSearch = "<<localSearchFlag<<"; OrderHeuristic = "<<orderHeuristic<<endl;

	//MAKE CANDIDATE SOLUTION STRUCTURES
	vector<vector<PIECE> > candSol, unplaced;
	vector<long> stockUsage;
		
	//DECLARE SOME CONTAINERS USED IN THE PROCEDURES (WE DON'T WANT TO DECLARE THESE ON EVERY ITERATION)
	vector<int> groupsToRemove;
	vector<PIECE> finalStock;

	//START THE TIMER
	clock_t clockStart = clock();
	clock_t clockFinish = clockStart + (timeLimit*CLOCKS_PER_SEC);
	
	//FORM THE INITIAL SOLUTION AND SEE HOW LONG IT TOOK
	makeInitSolution(candSol,stockUsage,finalStock,true);
	long CPUInitial = long((clock()-clockStart)/double(CLOCKS_PER_SEC)*1000.0);
	
	//Write details of initial solutions to the files
	binItsStream<<candSol.size()<<"\t0\n";
	binTimeStream<<candSol.size()<<"\t"<<CPUInitial<<"\n";
		
	//NOW DECIDE IF WE WANT TO RUN THE HILLCLIMBING ALGORITHM
	int i=0, currentNumGroups = candSol.size();
	bool stopRun = false;
	//Check to see if we should enter the main loop
	if(candSol.size() <= targetGroups) stopRun = true;
	else if(usingTL && usingIL && clockFinish < clock() && i>=itLimit) stopRun = true;
	else if(!usingTL && usingIL && i>=itLimit) stopRun = true;
	else if(usingTL && !usingIL && clockFinish < clock()) stopRun = true;

	//Some screen output
	if(verbosity>=1){
		cout<<"      It     Grp    Targ            MS    NumSubProbs\n"
			<<"      --     ---    ----            --    -----------\n"
			<<"       0"<<setw(8)<<candSol.size()<<setw(8)<<targetGroups<<setw(14)<<CPUInitial<<setw(15)<<TotalNumSubProbs<<endl;
	}

	//******************************************************************************************
	//Main Loop
	while(!stopRun){

		applySearch(candSol,stockUsage,unplaced,finalStock,groupsToRemove);

		i++;

		if(verbosity>=1)
			cout<<setw(8)<<i<<setw(8)<<candSol.size()<<setw(8)<<targetGroups<<setw(11)<<setw(14)<<long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000)<<setw(15)<<TotalNumSubProbs<<endl;

		if(candSol.size() < currentNumGroups) {
			currentNumGroups = candSol.size(); 
			binTimeStream<<currentNumGroups<<"\t"<<(int)(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000)<<"\n";
			binItsStream<<currentNumGroups<<"\t"<<i<<"\n";
		}
	
		//Check to see if we should end
		if(currentNumGroups <= targetGroups) stopRun = true;
		else if(usingTL && usingIL && clockFinish < clock() && i>=itLimit-1) stopRun = true;
		else if(!usingTL && usingIL && i>=itLimit-1) stopRun = true;
		else if(usingTL && !usingIL && clockFinish < clock()) stopRun = true;

	}
	//******************************************************************************************
	
	clockFinish = clock();
	long duration;
	duration = long(((clockFinish-clockStart)/double(CLOCKS_PER_SEC))*1000);

	//Check that the solution is valid (bug check)
	checkSolution(candSol,stockUsage);
	
	//Write solution for the visualiser
	prettyPrintOutputSolution(candSol,stockUsage);
	
	//Output stuff to screen if needed
	if(verbosity>=1){
		if(currentNumGroups == targetGroups) cout<<"Solution found that uses target number of bins"<<endl;
		cout<<"\n\nFinal Solution ("<<i+1<<" iterations):";
		prettyPrintSolution(candSol,stockUsage);
		prettyPrintSolutionProj(candSol, stockUsage);
		cout<<"\n\n----------------\nSummary\n----------------"<<endl;
		cout<<"TotalNumSubProbs = "<<TotalNumSubProbs<<endl;
		cout<<"BoundNoSolCnt    = "<<BoundNoSolCnt<<endl;
		cout<<"SuccGreedyCnt    = "<<SuccGreedyCnt<<endl;
		cout<<"SuccLSCnt        = "<<SuccLSCnt<<endl;
		cout<<"SuccEulerCnt     = "<<SuccEulerCnt<<endl;
		cout<<"FitNoSolCnt      = "<<FitNoSolCnt<<endl;
		cout<<"----------------"<<endl;
		cout <<"Time taken    = "<< duration <<" ms of CPU time"<<endl;
	}
	if(TotalNumSubProbs != SuccGreedyCnt + SuccEulerCnt + BoundNoSolCnt + SuccLSCnt + FitNoSolCnt)
		cout<<"Error: Some subproblems were unaccounted for\n";

	//Output Results and close output streams
	resultsLog<<currentNumGroups<<"\tX\t"<<duration<<"\t"<<i+1<<"\t"<<TotalNumSubProbs<<"\t"<<BoundNoSolCnt<<"\t"<<SuccGreedyCnt<<"\t"<<SuccLSCnt<<"\t"<<SuccEulerCnt<<"\t"<<FitNoSolCnt<<"\n";
	binTimeStream<<currentNumGroups-1<<"\tX\t"<<duration<<"\n";
	binItsStream<<currentNumGroups-1<<"\tX\t"<<i<<"\n";

	if(verbosity>=1)
		cout<<"Finishing Experiment with HillClimber; File = "<<filename<<"; SubProblemMethod = "<<subProbMethod<<"; LocalSearch = "<<localSearchFlag<<"; OrderHeuristic = "<<orderHeuristic<<endl;

}
