#include "trappack.h"

//Global Variables used for keeping track of how often different things were used.
long TotalNumSubProbs = 0;
long BoundNoSolCnt = 0;
long SuccGreedyCnt = 0;
long SuccEulerCnt = 0;
long SuccLSCnt = 0;
long FitNoSolCnt = 0;

//Global variables defining the size of the stocks
long STOCK_LEN = 4200;
long STOCK_WID = 2; //WARNING: THE STOCK WIDTH MUST BE 2
long STOCK_AREA = 8400; //WARNING: THIS MUST BE STOCK_LEN*STOCK_WID

//Global variables defining run parameters
int subProbMethod;
int verbosity;
int algType;
int localSearchFlag;
int orderHeuristic;
int repairHeuristic;

//Global variables defining problem paramters
int numItems;
int numTypes;
int tMin;
vector<SHAPE> shape;
vector<TYPE> types;

//Global Variables defining output streams
ofstream binTimeStream;
ofstream binItsStream;
ofstream costTimeStream;
ofstream costItsStream;
ofstream divStream;
ofstream unplStream;
ofstream resultsLog;

//Info output if different no parameters used
void usage() {
	cout<<"TRAPEZOID PACKING PROGRAM\n";
	cout<<"-------------------------\n";
	cout<<"USAGE:\n"
	<<"inputFile <filename>\n"
	<<"---- Optional Run Parameters -------------------------------\n"
	<<"-t  <int>         (Time limit. Default = 300s)\n"
	<<"-i  <int>         (Iteration limit. Default = undefined)\n"
	<<"-r  <int>         (Random seed. Default = 1)\n"
	<<"-k  <int>         (Target num. groups. Default = TMin, otherwise stops when target is met or when -t or -i is met)\n"
	<<"---- Optional Run Settings ---------------------------------\n"
	<<"-l  <int>         (Local search. Default = 1. If set to zero, no LS takes place; else -l rounds of LS are carried out per cycle)\n"
	<<"-a  <int>         (Algorithm. Default = 1.\n"
	<<"                          1: HillClimber;\n"
	<<"                          2: GA with GGA recombination;\n"
	<<"                          3: GA with GPX recombination;)\n"
	<<"                          4: GA with GPX recombination where whole groups are eliminated.)\n"
	<<"-s  <int>         (Subproblem algorithm. Default = 2.\n"
	<<"                          1: Greedy + Flips;\n"
	<<"                          2: Exact (Euler-Splice) Method.)\n"
	<<"-o  <int>         (Order Heuristic. Default = 1.\n"
	<<"                          1: Put items in random order\n"
	<<"                          2: Random order, but with all items >50% of capacity first)\n"
	<<"-g  <int>         (Genetic Repair Heuristic. Deafult = 1.\n"
	<<"                          1: Choose bin with common item at random;\n"
	<<"                          2: Choose bin with most common items;\n"
	<<"                          3: Choose bin with common item that has the most spare capacity.)\n"
	<<"-p  <int>         (Population size (GAs only). Default = 10.)\n"
	<<"-np <int>         (Number of parents per recombination (GAs only). Default = 2.)\n"
	<<"---- Optional Output Settings ------------------------------\n"
	<<"-v                (Verbosity. If present screen output is produced)\n"
	<<"-d                (Diversity. If present, diversity is output to a file and the screen at every iteration. Only used with EAs. Warning: Makes the program slower.)\n"
	<<"------------------------------------------------------------\n"
	<<"NOTES:\n"
	<<"If -t is set to a value then iterations (-i) is not used as a stopping criteria.\n"
	<<"If -i is set to a value then time (-t) is not used as a stopping criteria.\n"
	<<"If both -i and -t are set then the run ends when BOTH have been met.\n"
	<<"Order heuristic (-o) is used in the GAs only (for generating initial solutions and with the mutation operator).\n"
	<<"Genetic repair heuristic is only used with GA recombination. It is used as part of the set covering sub-problem for deciding which item types to delete from which bin during repair.\n"
	<<"Number of parents per recombination (-np) must be less than the population size defined by -p\n";
	exit(0);
}

int main(int argc, char *argv[]){

	if(argc <=1){
		usage();
	}

	//Determine run variables and set default values
	subProbMethod=2;
	verbosity=0;
	algType=1;
	localSearchFlag=1;
	orderHeuristic=1;
	repairHeuristic=1;
	int timeLimit = 0, itLimit = 0, seed = 1, targetGroups = -1, i, popSize = 10, numParents=2;
	bool usingTL = false, usingIL = false, stopRun = false, diversityOn=false;
	for (i=1; i<argc; i++) {
		if (strcmp("-t", argv[i]) == 0) {
		  timeLimit = atoi(argv[++i]);
		  usingTL = true;
		} else if (strcmp("-i", argv[i]) == 0 ) {
		  itLimit = atoi(argv[++i]);
		  usingIL = true;
		} else if (strcmp("-l", argv[i]) == 0 ) {
		  localSearchFlag = atoi(argv[++i]);
		} else if (strcmp("-a", argv[i]) == 0 ) {
		  algType = atoi(argv[++i]);
		} else if (strcmp("-r", argv[i]) == 0 ) {
		  seed = atoi(argv[++i]);
		} else if (strcmp("-k", argv[i]) == 0 ) {
		  targetGroups = atoi(argv[++i]);
		} else if (strcmp("-v", argv[i]) == 0 ) {
		  verbosity++;
		} else if (strcmp("-o", argv[i]) == 0 ) {
		  orderHeuristic = atoi(argv[++i]);
		} else if (strcmp("-g", argv[i]) == 0 ) {
		  repairHeuristic = atoi(argv[++i]);
		} else if (strcmp("-p", argv[i]) == 0 ) {
		  popSize = atoi(argv[++i]);
		} else if (strcmp("-np", argv[i]) == 0 ) {
		  numParents = atoi(argv[++i]);
		} else if (strcmp("-s", argv[i]) == 0 ) {
		  subProbMethod = atoi(argv[++i]);
		} else if (strcmp("-d", argv[i]) == 0 ) {
		  diversityOn = true;
		} else {
			//read in the problem file and costruct the relevant arrays.
			ifstream inStream; inStream.open(argv[1]); 
			if (inStream.fail()){ cout << "ERROR OPENING INPUT FILE"; exit(1); }
			int numOrders,seedIndex;
			inStream >> numItems >> numTypes >> numOrders >> seedIndex;
			shape.resize(numItems+1);
			types.resize(numTypes);
			readInputFile(inStream);
			inStream.close();
		}
	}

	//Sort out stopping criteria
	if(!usingTL && !usingIL){
		timeLimit = 300;
		usingTL = true;
	}

	//Set random seed
	srand(seed);

	//If target is undefined, set target groups to tMin
	if(targetGroups==-1) targetGroups = tMin;

	//If GGA is being used, number of parents must be 2
	if(algType==2 && numParents != 2){
		cout<<"GGA recombination must use 2 parents. Setting numParents = 2...\n";
		numParents = 2;
	}

	//If any GA is being used, the popSize must be >=2 and numParents must be <= popSize
	if(algType != 1 && (popSize < 2 || numParents > popSize)){
		cout<<"Error in GA input parameters. Ending...\n"; exit(1);
	}
	
	//SET UP SOME OUTPUT FILES----------------------------------------------------------------------
	binTimeStream.open("tEffort.txt");
	binItsStream.open("iEffort.txt");
	divStream.open("diversity.txt");
	unplStream.open("unplaced.txt");
	costTimeStream.open("tCostEffort.txt");
	costItsStream.open("iCostEffort.txt");
	resultsLog.open("resultsLog.txt", ios::app);
	if (costTimeStream.fail() || costItsStream.fail() || binTimeStream.fail() || binItsStream.fail() || divStream.fail() || unplStream.fail()){ cout << "Error opening one of the output files"; exit(1);}

	//OUTPUT RUN CONDITIONS
	resultsLog<<argv[1]<<"\t"<<numItems<<"\t"<<numTypes<<"\t"<<targetGroups<<"\t"<<tMin<<"\t"<<timeLimit*1000<<"\t"<<itLimit<<"\t"<<seed<<"\t"<<algType<<"\t"<<localSearchFlag<<"\t"<<subProbMethod<<"\t"<<orderHeuristic<<"\t"<<repairHeuristic<<"\t"<<diversityOn<<"\t";
	if(algType==1)resultsLog<<"-\t-\t";
	else resultsLog<<popSize<<"\t"<<numParents<<"\t";

	//NOW START THE ALGORITHM
	if(algType == 1){
		runHillClimber(targetGroups,usingTL,usingIL,itLimit,timeLimit,argv[1]);
	} else { 
		runGA(targetGroups,usingTL,usingIL,itLimit,timeLimit,argv[1],diversityOn,popSize,numParents);
	}

	//ALGORITHM FINISHED, CLOSE STREAMS AND END
	binItsStream.close(); 
	binTimeStream.close();
	costItsStream.close(); 
	costTimeStream.close();
	divStream.close();
	resultsLog.close();
	unplStream.close();
	return 0;
}


